package jp.ne.sakura.sabowl.gpsnmeajp.quartzpaydemo;

import android.app.Activity;
import android.content.BroadcastReceiver;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.nfc.NfcAdapter;
import android.nfc.cardemulation.NfcFCardEmulation;
import android.support.v4.content.LocalBroadcastManager;
import android.util.Log;

/**
 * <br>HCE-FのコマンドをFeliCa Lite相当で処理し，コールバックする．
 * <br>Activity側のクラス
<pre>
 LazyNfcFHCELite

 zlib License
 Copyright (c) 2018 GPS_NMEA_JP

 This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
 Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
 1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
 2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
 3. This notice may not be removed or altered from any source distribution.
 </pre>

 * @author gpsnmeajp
 */
public class LazyNfcFHCELite {
    //基本情報
    private String debugMsgTitle = "LazyNfcFHCELite";

    private Activity activity = null;
    private NfcAdapter nfcAdapter;
    private NfcFCardEmulation nfcFCardEmulation;
    private ComponentName componentName;

    //コールバック保持
    private onHCEFAccessCallback HCEFAccessCallback = null;

    //FeliCa Lite定数
    private final int PACKET_LENGTH_POS = 0;
    private final int PACKET_LENGTH_SIZE = 1;
    private final int COMMAND_TYPE_POS = 1;
    private final int COMMAND_TYPE_SIZE = 1;
    private final int IDM_POS = 2;
    private final int IDM_SIZE = 8;
    private final int RESPONSE_SIZE = 1;
    private final int STATUS_FLAG1_SIZE = 1;
    private final int STATUS_FLAG2_SIZE = 1;
    private final int BLOCK_NUM_SIZE = 1;
    private final int BLOCK_SIZE = 16;
    private final int ONE_BLOCK = 1;

    private final byte CMD_POLLING = (byte)0x04;
    private final byte RES_POLLING = (byte)0x05;
    private final byte CMD_READ_WITHOUT_ENCRYPTION = (byte)0x06;
    private final byte RES_READ_WITHOUT_ENCRYPTION = (byte)0x07;
    private final byte CMD_WRITE_WITHOUT_ENCRYPTION = (byte)0x08;
    private final byte RES_WRITE_WITHOUT_ENCRYPTION = (byte)0x09;

    private final byte STATUS_FLAG1_SUCCESS = (byte)0x00;
    private final byte STATUS_FLAG1_FAILED = (byte)0xFF;
    private final byte STATUS_FLAG1_FELICA_LITE_ERROR = (byte)0x01;

    private final byte STATUS_FLAG2_SUCCESS = (byte)0x00;
    private final byte STATUS_FLAG2_READ_ERROR = (byte)0x70;
    private final byte STATUS_FLAG2_READONLY = (byte)0xA8;
    private final byte STATUS_FLAG2_NEED_AUTH = (byte)0xB1;
    private final byte STATUS_FLAG2_SERVICE_NUM_ERROR = (byte)0xA1;
    private final byte STATUS_FLAG2_BLOCK_NUM_ERROR = (byte)0xA2;
    private final byte STATUS_FLAG2_SERVICE_CODE = (byte)0xA6;
    private final byte STATUS_FLAG2_ACCESS_MODE = (byte)0xA7;

    private final byte SERVICE_CODE_READ_ONLY = 0x000B;
    private final byte SERVICE_CODE_READ_WRITE = 0x0009;

    //コンストラクタ(各変数の初期化)
    public LazyNfcFHCELite() {}
    public LazyNfcFHCELite(Activity activity,String packageName, String HCEService) {
        register(activity,packageName,HCEService);
    }

    /**
     * この端末がHCE-Fをサポートしているかを返す
     * @param activity 現在のActivity
     * @return true:サポートしている, false:サポートしていない
     */
    public static boolean isHCESupported(Activity activity)
    {
        PackageManager pm = activity.getPackageManager();
        return pm.hasSystemFeature(PackageManager.FEATURE_NFC_HOST_CARD_EMULATION_NFCF);
    }

    /**
     * Activityの登録と各インスタンスの生成， LocalBroadcastManagerへの登録を行う．
     * ActivityのonCreateで呼び出すのがちょうどよい．
     *
     * @param activity 現在のActivity
     * @param packageName 文字列．現在のパッケージ名を入れる．例:jp.ne.sakura.sabowl.gpsnmeajp.intent_hce
     * @param LazyNfcFHCELiteProxy 文字列．Proxyサービスのフル名を入れる．例:jp.ne.sakura.sabowl.gpsnmeajp.intent_hce.LazyNfcFHCELiteProxy
     */
    public void register(Activity activity,String packageName, String LazyNfcFHCELiteProxy)
    {
        this.activity = activity;
        componentName = new ComponentName(packageName,LazyNfcFHCELiteProxy);
        nfcAdapter = NfcAdapter.getDefaultAdapter(activity);
        nfcFCardEmulation = NfcFCardEmulation.getInstance(nfcAdapter);

        //HCE-Fサービスからの情報受信用
        LocalBroadcastManager.getInstance(activity).registerReceiver(mProcessNfcFPacketReceiver,
                new IntentFilter("LazyNfcFHCELite::processNfcFPacket"));
        LocalBroadcastManager.getInstance(activity).registerReceiver(mOnDeactivatedReceiver,
                new IntentFilter("LazyNfcFHCELite::onDeactivated"));
    }

    /**
     * Activityの登録解除と各インスタンスの破棄， LocalBroadcastManagerへの登録解除を行う．
     */
    public void unregister()
    {
        this.activity = null;
        componentName = null;
        nfcAdapter = null;
        nfcFCardEmulation = null;

        //HCE-Fサービスからの情報受信を解除
        LocalBroadcastManager.getInstance(activity).unregisterReceiver(mProcessNfcFPacketReceiver);
        LocalBroadcastManager.getInstance(activity).unregisterReceiver(mOnDeactivatedReceiver);
    }

    /**
     * システムにサービスの存在を知らせ，HCE-Fサービスを有効化する．
     * ActivityのonResumeで呼び出すことが推奨されている．
     * なお，NFC-Fの制約上，アプリケーションがフォアグラウンドから外れると強制的に無効化される．
     */
    public void enable()
    {
        Log.d(debugMsgTitle,"Enable service");
        nfcFCardEmulation.enableService(activity, componentName);
    }

    /**
     * HCE-Fサービスを無効化する．
     * ActivityのonPauseで呼び出すことが推奨されている．
     * なお，NFC-Fの制約上，アプリケーションがフォアグラウンドから外れると強制的に無効化される．
     */
    public void disable()
    {
        Log.d(debugMsgTitle,"Disable service");
        nfcFCardEmulation.disableService(activity);
    }

    /**
     * FeliCa Lite相当のNFCパケットを受信した際に呼び出されるコールバックを登録する
     *
     * @param callback onHCEFAccessCallbackインターフェースにもとづいて生成されたインスタンス
     */
    public void registerCallback(final onHCEFAccessCallback callback){
        this.HCEFAccessCallback = callback;
    }

    /**
     * コールバックを登録解除する
     */
    public void unregisterCallback(){
        this.HCEFAccessCallback = null;
    }

    /**
     * FeliCa Lite相当のNFCパケットを受信した際に呼び出されるコールバックのインターフェース
     */
    interface onHCEFAccessCallback {

        /**
         * ReadWithoutEncryptionコマンド時に呼び出される．
         * @param  serviceCode サービスコード(0-65535)
         * @param  address ブロックアドレス(0-65535)
         * @return 16バイトの配列か，null．(nullの場合は送信元にサービスコードエラーで応答する)
         */
        byte[] onRead(int serviceCode, int address);

        /**
         * WriteWithoutEncryptionコマンド時に呼び出される．
         * @param  serviceCode サービスコード(0-65535)
         * @param  address ブロックアドレス(0-65535)
         * @param  data 書き込まれたデータ(16Byte)
         * @return trueで書き込み成功応答，falseはサービスコードエラーで応答する．
         */
        //コマンド時に呼び出し．trueで成功応答．falseでエラー応答
        boolean onWrite(int serviceCode, int address, byte data[]);

        /**
         * RF切断時に呼び出される．
         * @param  reason 切断理由コード
         */
        void onDeactivate(int reason);
    }

    //--------------------------------------

    /**
     * ProcessNfcFPacketを受信，再現．
     * コマンド解析の結果をsendResponsePacketで遅延返却している．
     */
    private BroadcastReceiver mProcessNfcFPacketReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            byte commandPacket[] = intent.getByteArrayExtra("packet");
            byte responsePacket[] = processNfcFPacket(commandPacket);
            sendResponsePacket(responsePacket);
        }
    };

    /**
     * OnDeactivatedを受信，再現
     */
    private BroadcastReceiver mOnDeactivatedReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            int reason = intent.getIntExtra("reason",0);
            onDeactivated(reason);
        }
    };

    /**
     * サービス側へsendResponsePacketの実行を依頼する．
     *
     * 無応答としていたパケットのレスポンスを後から送信するためのメソッド．
     * サービス側では必ず無応答としているため，このメソッドを用いて応答を返却する．
     * リーダライター側のタイムアウトより早く応答する必要がある．
     *
     * @param responsePacket NFC応答パケット(nullの場合は送信しない)
     */
    public void sendResponsePacket(byte responsePacket[])
    {
        if(responsePacket != null) {
            Intent responceIntent = new Intent("LazyNfcFHCELite::sendResponsePacket");
            responceIntent.putExtra("packet", responsePacket);
            LocalBroadcastManager.getInstance(activity).sendBroadcast(responceIntent);
        }
    }

    /**
     * RF切断時に呼び出される．
     * コールバックonDeactivateを呼び出す
     * @param  reason 切断理由コード
     */
    private void onDeactivated(int reason)
    {
        Log.d(debugMsgTitle,"onDeactivated");
        if(HCEFAccessCallback != null) {
            HCEFAccessCallback.onDeactivate(reason);
        }
    }

    /**
     * システムから受信したNFCパケットを処理．
     * サイズチェックを行って，FeliCaコマンドを解析して各コマンド処理メソッドをコール
     * @param commandPacket システムから受信したNFCパケット
     * @return 応答パケット
     */
    private byte[] processNfcFPacket(byte[] commandPacket)
    {
        Log.d(debugMsgTitle,"processNfcFPacket");

        //必須情報より小さいならエラー
        if (commandPacket.length < (PACKET_LENGTH_SIZE+COMMAND_TYPE_SIZE+IDM_SIZE)) {
            Log.e(debugMsgTitle,"processNfcFPacket: Packet size too short");
            return null;
        }

        //1バイト目をチェックして分岐．FeliCa Liteに相当
        byte commandType = commandPacket[COMMAND_TYPE_POS];
        switch(commandType)
        {
            case CMD_POLLING: return null;//stub
            case CMD_WRITE_WITHOUT_ENCRYPTION: return WriteWithoutEncryption(commandPacket);//stub
            case CMD_READ_WITHOUT_ENCRYPTION: return ReadWithoutEncryption(commandPacket);//stub
            default: return null; //該当コマンドなし
        }
    }

    /**
     * Write without Encryptionコマンドとしてパケットを解析し，各種情報を取得．
     * コールバックonWriteを呼び出す．存在しなければサービス番号エラーを返却する．
     * エラーがあればFeliCaカードコマンドのエラー形式で返却する．
     *
     * @param commandPacket システムから受信したNFCパケット
     * @return 応答パケット(Write without encryption形式)
     */
    private byte[] WriteWithoutEncryption(byte[] commandPacket)
    {
        Log.d(debugMsgTitle,"WriteWithoutEncryption");
        //response
        int len = PACKET_LENGTH_SIZE + RESPONSE_SIZE + IDM_SIZE + STATUS_FLAG1_SIZE + STATUS_FLAG2_SIZE;
        byte[] responsePacket = new byte[len];
        responsePacket[PACKET_LENGTH_POS] = (byte)len;
        responsePacket[COMMAND_TYPE_POS] = RES_WRITE_WITHOUT_ENCRYPTION;
        responsePacket[len-2] = STATUS_FLAG1_SUCCESS;
        responsePacket[len-1] = STATUS_FLAG2_SUCCESS;

        //IDmをセット
        responsePacket = setIDmToPacket(responsePacket,getIDm(commandPacket));

        //analyze
        int serviceNum = commandPacket[10]&0xFF;
        if(serviceNum != 1)
        {
            responsePacket[len-2] = STATUS_FLAG1_FAILED;
            responsePacket[len-1] = STATUS_FLAG2_SERVICE_NUM_ERROR;
            Log.d(debugMsgTitle,"SERVICE_NUM_ERROR");
            return responsePacket; //ERROR RES
        }

        int serviceCode = (commandPacket[11]&0xFF) | ((commandPacket[12]&0xFF)<<8);
        Log.d(debugMsgTitle,String.format("Service code : %04X,",serviceCode));

        int blockNum = commandPacket[13]&0xFF;
        if(blockNum != 1)
        {
            responsePacket[len-2] = STATUS_FLAG1_FAILED;
            responsePacket[len-1] = STATUS_FLAG2_BLOCK_NUM_ERROR;
            Log.d(debugMsgTitle,"BLOCK_NUM_ERROR");
            return responsePacket; //ERROR RES
        }

        int blockAddress = -1;
        int blockHeadPos = -1;
        int blocklist_1 = commandPacket[14]&0xFF;
        if(blocklist_1 == 0x80)
        {
            //1byte
            blockAddress = commandPacket[15]&0xFF;
            blockHeadPos = 16;
        }else if(blocklist_1 == 0x00)
        {
            //2byte
            blockAddress = (commandPacket[15]&0xFF) | ((commandPacket[16]&0xFF)<<8);
            blockHeadPos = 17;
        }else{
            responsePacket[len-2] = STATUS_FLAG1_FAILED;
            responsePacket[len-1] = STATUS_FLAG2_ACCESS_MODE;
            Log.d(debugMsgTitle,"SERVICE_NUM_ERROR");
            return responsePacket; //ERROR RES
        }

        Log.d(debugMsgTitle,String.format("Block Address : %04X,",blockAddress));

        byte data[] = new byte[BLOCK_SIZE];
        System.arraycopy(commandPacket, blockHeadPos,data , 0, BLOCK_SIZE);

        String debug="";
        for(int i=0;i<data.length;i++)
        {
            debug += String.format("%02X,",data[i]);
        }
        Log.d(debugMsgTitle,debug);

        if(HCEFAccessCallback != null) {
            if(!HCEFAccessCallback.onWrite(serviceCode,blockAddress,data))
            {
                responsePacket[len-2] = STATUS_FLAG1_FELICA_LITE_ERROR;
                responsePacket[len-1] = STATUS_FLAG2_SERVICE_CODE;
                Log.d(debugMsgTitle,"STATUS_FLAG2_SERVICE_CODE");
                return responsePacket; //ERROR RES
            }
        }
        Log.d(debugMsgTitle,"SUCCESS");
        return responsePacket; //SUCCESS
    }

    /**
     * Read without Encryptionコマンドとしてパケットを解析し，各種情報を取得．
     * コールバックonReadを呼び出す．存在しなければサービス番号エラーを返却する．
     * エラーがあればFeliCaカードコマンドのエラー形式で返却する．
     *
     * @param commandPacket システムから受信したNFCパケット
     * @return 応答パケット(Read without encryption形式)
     */
    private byte[] ReadWithoutEncryption(byte[] commandPacket)
    {
        Log.d(debugMsgTitle,"ReadWithoutEncryption");
        //response
        int len = PACKET_LENGTH_SIZE + RESPONSE_SIZE + IDM_SIZE + STATUS_FLAG1_SIZE + STATUS_FLAG2_SIZE + BLOCK_NUM_SIZE + BLOCK_SIZE;
        int headLen = PACKET_LENGTH_SIZE + RESPONSE_SIZE + IDM_SIZE + STATUS_FLAG1_SIZE + STATUS_FLAG2_SIZE;
        byte[] responsePacket = new byte[len];
        responsePacket[PACKET_LENGTH_POS] = (byte)len;
        responsePacket[COMMAND_TYPE_POS] = RES_READ_WITHOUT_ENCRYPTION;
        responsePacket[headLen-2] = STATUS_FLAG1_SUCCESS;
        responsePacket[headLen-1] = STATUS_FLAG2_SUCCESS;
        responsePacket[headLen] = ONE_BLOCK;

        //IDmをセット
        responsePacket = setIDmToPacket(responsePacket,getIDm(commandPacket));

        //analyze
        int serviceNum = commandPacket[10]&0xFF;
        if(serviceNum != 1)
        {
            responsePacket[len-2] = STATUS_FLAG1_FAILED;
            responsePacket[len-1] = STATUS_FLAG2_SERVICE_NUM_ERROR;
            Log.d(debugMsgTitle,"SERVICE_NUM_ERROR");
            return responsePacket; //ERROR RES
        }

        int serviceCode = (commandPacket[11]&0xFF) | ((commandPacket[12]&0xFF)<<8);
        Log.d(debugMsgTitle,String.format("Service code : %04X,",serviceCode));

        int blockNum = commandPacket[13]&0xFF;
        if(blockNum != 1)
        {
            responsePacket[len-2] = STATUS_FLAG1_FAILED;
            responsePacket[len-1] = STATUS_FLAG2_BLOCK_NUM_ERROR;
            Log.d(debugMsgTitle,"BLOCK_NUM_ERROR");
            return responsePacket; //ERROR RES
        }

        int blockAddress = -1;
        int blocklist_1 = commandPacket[14]&0xFF;
        if(blocklist_1 == 0x80)
        {
            //1byte
            blockAddress = commandPacket[15]&0xFF;
        }else if(blocklist_1 == 0x00)
        {
            //2byte
            blockAddress = (commandPacket[15]&0xFF) | ((commandPacket[16]&0xFF)<<8);
        }else{
            responsePacket[len-2] = STATUS_FLAG1_FAILED;
            responsePacket[len-1] = STATUS_FLAG2_ACCESS_MODE;
            Log.d(debugMsgTitle,"SERVICE_NUM_ERROR");
            return responsePacket; //ERROR RES
        }

        Log.d(debugMsgTitle,String.format("Block Address : %04X,",blockAddress));

        byte data[] = null;
        if(HCEFAccessCallback != null) {
            data = HCEFAccessCallback.onRead(serviceCode,blockAddress);
        }
        if(data == null || data.length != 16)
        {
            responsePacket[len-2] = STATUS_FLAG1_FAILED;
            responsePacket[len-1] = STATUS_FLAG2_READ_ERROR;
            Log.d(debugMsgTitle,"SERVICE_NUM_ERROR");
            return responsePacket; //ERROR RES
        }
        System.arraycopy(data, 0,responsePacket , headLen + BLOCK_NUM_SIZE, BLOCK_SIZE);

        String debug="";
        for(int i=0;i<responsePacket.length;i++)
        {
            debug += String.format("%02X,",responsePacket[i]);
        }
        Log.d(debugMsgTitle,debug);
        Log.d(debugMsgTitle,"SUCCESS");

        return responsePacket;
    }

    /**
     * パケットを解析し，IDmを取得する．
     *
     * @param commandPacket システムから受信したNFCパケット
     * @return IDm(8byte)
     */
    private byte[] getIDm(byte[] commandPacket)
    {
        byte[] IDm = new byte[IDM_SIZE];
        System.arraycopy(commandPacket,IDM_POS, IDm, 0, IDM_SIZE);

        return IDm;
    }

    /**
     * パケットのIDm部分に，引数のIDmを上書きする
     *
     * @param packet システムから受信したNFCパケット
     * @param IDm IDm(8byte)
     * @return IDmが上書きされたNFCパケット
     */
    private  byte[] setIDmToPacket(byte[] packet,byte[] IDm)
    {
        System.arraycopy(IDm, 0, packet, IDM_POS, IDM_SIZE);// NFCID2
        return packet;
    }


}
